// Copyright 2006-2007 all rights reserved; see LICENSE file for BSD-style license

package edu.berkeley.sbp;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.Sequence.Pos;
import edu.berkeley.sbp.Sequence.Pos;
import java.util.*;

final class ResultNode extends Node {

    private Forest.Many f = new Forest.Many();
    //private HashSet<StateNode> predecessors = new HashSet<StateNode>();
    //private HashSet<StateNode> successors = new HashSet<StateNode>();
    private FastSet<StateNode> predecessors = new FastSet<StateNode>();
    private FastSet<StateNode> successors = new FastSet<StateNode>();
    private boolean destroyed = false;
    private boolean primordeal;
    private int usedByNonDoomedNode = 0;
    private Pos reduction;
    private GSS.Phase predPhase;

    public boolean predecessorsContains(StateNode n) {
        return predecessors.contains(n);
    }
    public Pos reduction() { return reduction; }
    public void merge(Forest newf) {
        this.f.merge(newf);
        /*
        if (predecessors.contains(pred)) return;
        addPred(pred);
        if (fromEmptyReduction) return;
        n.state().invokeReductions(n.phase().getToken(), n, this);        
        */
    }

    public boolean noSuccessors() { return successors.size()==0; }

    public GSS.Phase phase() { return predPhase; }
    public Forest getForest() { return f; }
    public Iterable<StateNode> getPreds() { return predecessors; }
    public void addSucc(StateNode succ) {
        if (successors.contains(succ)) return;
        successors.add(succ);
        usedByNonDoomedNode += succ.state().doomed ? 0 : 1;
        if (predecessors.size() > 1) throw new Error();
    }
    public void removeSucc(StateNode succ) {
        if (!successors.contains(succ)) return;
        successors.remove(succ);
        usedByNonDoomedNode -= succ.state().doomed ? 0 : 1;
        check();
    }

    public boolean usedByAnyNode() { return successors.size() > 0; }
    public boolean usedByNonDoomedNode() { return usedByNonDoomedNode > 0; }

    public String toString() { return super.toString()+"->"+predPhase; }

    public void check() {
        if (successors.size()==0) destroy();
        else if (predecessors.size()==0) destroy();
    }
    public void destroy() {
        if (destroyed) return;
        if (primordeal) return;  // never destroy the "primordeal" result
        destroyed = true;
        while(predecessors.size() > 0)
            for(StateNode pred : predecessors) {
                removePred(pred);
                pred.removeSucc(this);
                break;
            }
        predecessors = null;
        while(successors.size() > 0)
            for(StateNode succ : successors) {
                removeSucc(succ);
                succ.removeResult(this);
                break;
            }
        successors = null;
    }

    public void removePred(StateNode pred) {
        if (!predecessors.contains(pred)) return;
        predecessors.remove(pred);
        check();
    }

    public void addPred(StateNode pred) {
        if (predPhase==null) predPhase = pred.phase();
        if (predPhase != pred.phase()) throw new Error();
        predecessors.add(pred);
        pred.addSucc(this);
        if (predecessors.size() > 1) throw new Error();
    }
        
    public ResultNode() {
        this(null, null, null);
        this.primordeal = true;
    }
    public ResultNode(Forest f, Pos reduction, StateNode pred) {
        this.f.merge(f);
        this.reduction = reduction;
        if (pred != null) addPred(pred);
    }


}